#-*- coding:utf-8 -*-
from calendar import EPOCH
import os
from pickletools import optimize
import sys
import pandas as pd
import numpy as np
from sklearn.metrics import accuracy_score, precision_score, recall_score
import torch, gc
from torch.utils.data import DataLoader
import transformers
from torch.utils.data import TensorDataset, DataLoader, RandomSampler, SequentialSampler
import transformers
import keras_preprocessing.sequence
from sklearn.model_selection import train_test_split
from transformers import AdamW
from transformers import get_linear_schedule_with_warmup
from seqeval.metrics import accuracy_score
from tqdm import trange
import re
import evaluation
import wandb

tokenizer = transformers.BertTokenizer.from_pretrained("UWB-AIR/Czert-B-base-cased")
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.cuda.empty_cache()
MAX_LEN = 512
BATCH_SIZE = 10
EPOCHS = 10

"""
Sort names of docs (same for UNIX and WINDOWS)

Arguments:
    data: List to sort
"""
def sorted_alphanumeric(data):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ] 
    return sorted(data, key=alphanum_key)


"""
loads keywords and documents

Arguments:
    docs: Paths to individual files with documents
    keys: Paths to individual files with keywords

Returns:
    d: dataframe contains whole article in one column and 1 or 0 (keyword or not) in another
"""
def preprocessing(d_docs, d_key, docs, keys):
    words_array = []

    # load articles
    for i in range(0, len(docs)):
        file = open(d_docs + docs[i], encoding="utf8")
        line = file.read().replace("\n", " ").lower()
        file.close()
        words_array.append(line.split())

    assigned_kw = []
    # load keywords
    for i in range(0, len(keys)):
        file = open(d_key + keys[i], encoding="utf8")
        assigned_kw.append(file.read().splitlines())
        file.close()

    
    labels = []
    tmp = []
    # keyword = 1   non-keyword = 0
    for i in range(0, len(words_array)):
        for j in range(0, len(words_array[i])):
            if(words_array[i][j] in assigned_kw[i]):
                tmp.append(1)
            else:
                tmp.append(0)
        labels.append(tmp)
        tmp = []


    data_tuples = list(zip(words_array, labels))
    d = pd.DataFrame(data_tuples, columns=['article','label'])
    return d



"""
loads validation data (only if evaluation file is created already)

Arguments:
    dataset: Path to dataset

Returns:
    text_array:  List of whole articles
    lens:        List of articles lengths
    words_array: List of words of each article
"""
def load_val_data(dataset):
    if os.path.exists(dataset + "val/docs") == True:
        docs = sorted_alphanumeric(os.listdir(dataset + "val/docs"))
        folder = dataset + "val/docs/"
    else:
        docs = sorted(os.listdir(dataset + "docs"), key=len)
        folder = dataset + "docs/"

    text_array = []
    lens = []
    words_array = []
    for i in range(0, len(docs)):
        file = open(folder + docs[i], encoding="utf8")
        line = file.read().replace("\n", " ").lower()
        file.close()
        words = re.sub("[^\w]", " ",  line).split()
        words_array.append(words)
        lens.append(len(words))
        tmp = ' '.join([str(elem) for elem in words])
        text_array.append(tmp)
    return text_array, lens, words_array



"""
Computes acurracy, precission, recall and f score and create a eval file

Arguments:
    model: BERT-like model
    dataset: Path to dataset
    name: name of the output eval file

Returns:
    a: acurracy
    p: precission
    r: recall
    f: f score
"""
def eval_gold(model, dataset, name):
    if os.path.exists(name + "_eval.tsv") == True:
        os.remove(name + "_eval.tsv")

    val_data, lens, words_array = load_val_data(dataset)
    for i in range(0, len(val_data)):
        new_tokens, new_labels = [], []
        tokenized_sent = tokenizer.encode(val_data[i])
        original_len = lens[i]

        # crop to 512
        if(len(tokenized_sent) > MAX_LEN):
            tokenized_sent = tokenized_sent[:MAX_LEN]
        input_ids = torch.tensor([tokenized_sent]).cuda()
        with torch.no_grad():
            output = model(input_ids)
        
        label_indices = np.argmax(output[0].to('cpu').numpy(), axis=2)
        tokens = tokenizer.convert_ids_to_tokens(input_ids.to('cpu').numpy()[0])

        for token, label_idx in zip(tokens, label_indices[0]):
            if token == "[CLS]" or token == "[SEP]":
                continue
            if token.startswith("##"):
                new_tokens[-1] = new_tokens[-1] + token[2:]
            else:
                new_labels.append(label_idx)
                new_tokens.append(token)

        if original_len > len(new_labels):
            for j in range(len(new_labels), original_len):
                new_tokens.append("[CROPPED]")
                new_labels.append(0)
        new_tokens.append("###")
        new_labels.append("###")
		
        data_tuples = list(zip(new_tokens, new_labels))
        new_df = pd.DataFrame(data_tuples, columns=['WORD', 'KEYWORD'])
        rslt_df = new_df[new_df['KEYWORD'] == 1]

        keywords = rslt_df['WORD'].tolist()
        tmp = []
        text = []
        for art_w in range(0, len(words_array[i])):
            text.append(words_array[i][art_w])
            if(words_array[i][art_w] in keywords):
                tmp.append(1)
            else:
                tmp.append(0)
        tmp.append("###")
        text.append("###")

        data_tuples = list(zip(text, tmp))
        d = pd.DataFrame(data_tuples, columns=['WORD', 'KEYWORD'])
        d.to_csv(name + "_eval.tsv", sep="\t", index=False, mode='a')
    
    if not os.path.exists(dataset + "eval.tsv"):
        print("WARNING: Evaluation file does not exists!")
        return None, None, None, None

    eval_file = dataset + "eval.tsv"
    eval_df = pd.read_csv(eval_file, sep='\t')
    eval_vector = eval_df['KEYWORD'].tolist()

    pred_file = name + "_eval.tsv"
    pred_df = pd.read_csv(pred_file, sep='\t')
    pred_vector = pred_df['KEYWORD'].tolist()
    a, p, r, f = evaluation.evaluation(eval_vector, pred_vector)
    return a, p, r, f


"""
Tokenizes words and keep labels mapping on individual tokens

Arguments:
    art: one article
    text_labels: labels

Returns:
    tokenized_art: tokenized article
    labels: labels
"""
def tokenize_and_preserve_labels(art, text_labels):
    tokenized_art = []
    labels = []

    for word, label in zip(art, text_labels):

        # Tokenize the word and count # of subwords the word is broken into
        tokenized_word = tokenizer.tokenize(word)
        n_subwords = len(tokenized_word)

        # Add the tokenized word to the final tokenized word list
        tokenized_art.extend(tokenized_word)

        # Add the same label to the new list of labels `n_subwords` times
        labels.extend([label] * n_subwords)
    return tokenized_art, labels


"""
Get tokenized data and labels

Arguments:
    data: dataframe with a structure like ARTICLE in first column and KEYWORDS in second

Returns:
    input_ids: ids of each tokens
    labels_padd: labels
    attention_masks: attention masks
"""
def get_tokenized_data(data):
    articles = data['article']
    labels = data['label']


    tokenized_texts_and_labels = [
    tokenize_and_preserve_labels(art, labs)
    for art, labs in zip(articles, labels)
    ]

    tokenized_texts = [token_label_pair[0] for token_label_pair in tokenized_texts_and_labels]
    labels = [token_label_pair[1] for token_label_pair in tokenized_texts_and_labels]

    input_ids = keras_preprocessing.sequence.pad_sequences([tokenizer.convert_tokens_to_ids(txt) for txt in tokenized_texts],
                          maxlen=MAX_LEN, dtype="long", value=0.0,
                          truncating="post", padding="post")

    labels_padd = keras_preprocessing.sequence.pad_sequences(labels,
                     maxlen=MAX_LEN, value=0, padding="post",
                     dtype="long", truncating="post")

    attention_masks = [[float(i != 0.0) for i in ii] for ii in input_ids]

    return input_ids, labels_padd, attention_masks



"""
Fine tune BERT like model

Arguments:
    input_ids: ids of tokens
    labels_padd: labels
    att_mask: attentoin mask
    dataset: path to dataset

Returns:
    model: fine tuned model
    f_score: final f score

"""
def fine_tuning(input_ids, labels_padd, att_mask, opt, dataset):
    name = "model_" + opt

    tr_inputs, val_inputs, tr_labels, val_labels = train_test_split(input_ids, labels_padd, random_state=20, test_size=0.2)
    tr_masks, val_masks, _, _ = train_test_split(att_mask, input_ids, random_state=20, test_size=0.2)
    tag_values = [0, 1]

    tr_steps = 0
    val_steps = 0

    tr_inputs = torch.tensor(tr_inputs)
    val_inputs = torch.tensor(val_inputs)
    tr_labels = torch.tensor(tr_labels)
    val_labels = torch.tensor(val_labels)
    tr_masks = torch.tensor(tr_masks)
    val_masks = torch.tensor(val_masks)

    train_data = TensorDataset(tr_inputs, tr_masks, tr_labels)
    train_sampler = RandomSampler(train_data)
    train_dataloader = DataLoader(train_data, sampler=train_sampler, batch_size=BATCH_SIZE)
    valid_data = TensorDataset(val_inputs, val_masks, val_labels)
    valid_sampler = SequentialSampler(valid_data)
    valid_dataloader = DataLoader(valid_data, sampler=valid_sampler, batch_size=BATCH_SIZE)

    model = transformers.BertForTokenClassification.from_pretrained(
    "UWB-AIR/Czert-B-base-cased",
    hidden_dropout_prob = 0.2,
    num_labels=2,
    output_attentions = False,
    output_hidden_states = False
    )

    model.to(device)

    # choose optimizer
    if(opt == "adam"):
        optimizer = torch.optim.Adam(params=model.parameters(), lr=3e-5)

    elif(opt == "adamw_1"):
        param_optimizer = list(model.classifier.named_parameters())
        optimizer_grouped_parameters = [{"params": [p for n, p in param_optimizer]}]
        optimizer = AdamW(
            optimizer_grouped_parameters,
            lr=3e-5,
            eps=1e-8
        )

    elif(opt == "adamw_2"):
        param_optimizer = list(model.named_parameters())
        no_decay = ['bias', 'gamma', 'beta']
        optimizer_grouped_parameters = [
            {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
            'weight_decay_rate': 0.01},
            {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)],
            'weight_decay_rate': 0.0}
        ]

        optimizer = AdamW(
            optimizer_grouped_parameters,
            lr=3e-5,
            eps=1e-8
        )

    elif(opt == "sgd"):
        optimizer = torch.optim.SGD(params=model.parameters(), lr=3e-5)
        
    
    max_grad_norm = 1.0

    # Total number of training steps is number of batches * number of epochs
    total_steps = len(train_dataloader) * EPOCHS

    # Create the learning rate scheduler.
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=0,
        num_training_steps=total_steps
    )

    loss_values, validation_loss_values = [], []

    torch.cuda.empty_cache()
    for _ in trange(EPOCHS, desc="Epoch"):
    # ========================================
    #               Training
    # ========================================
        model.train()
        total_loss = 0

        # Training loop
        for step, batch in enumerate(train_dataloader):
            batch = tuple(t.to(device) for t in batch)
            b_input_ids, b_input_mask, b_labels = batch

            b_labels = b_labels.type(torch.LongTensor).to(device)
            model.zero_grad()

            outputs = model(b_input_ids, attention_mask=b_input_mask, labels=b_labels)
            # get the loss
            loss = outputs[0]
            #print(loss)
            loss.backward()
            total_loss += float(loss.item())
            # Clip the norm of the gradient
            # This is to help prevent the "exploding gradients" problem.
            torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=max_grad_norm)
            optimizer.step()
            scheduler.step()
            wandb.log({       
            "Train Loss": loss.item(),
            "Train steps": tr_steps
            })
            tr_steps += 1

        avg_train_loss = total_loss / len(train_dataloader)
        print("Average train loss: {}".format(avg_train_loss))
        wandb.log({       
            "Average train Loss": avg_train_loss
            })
        loss_values.append(avg_train_loss)


        # ========================================
        #               Validation
        # ========================================
        model.eval()
        # Reset the validation loss for this epoch.
        eval_loss = 0
        true_labels, predictions = [],[]
        for step, batch in enumerate(valid_dataloader):
            batch = tuple(t.to(device) for t in batch)
            b_input_ids, b_input_mask, b_labels = batch

            b_labels = b_labels.type(torch.LongTensor).to(device)

            with torch.no_grad():
                outputs = model(b_input_ids, attention_mask=b_input_mask, labels=b_labels)

            logits = outputs[1].detach().cpu().numpy()
            label_ids = b_labels.to('cpu').numpy()
            # Calculate the accuracy for this batch of test sentences.
            eval_loss += float(outputs[0].mean().item())
            predictions.extend([list(p) for p in np.argmax(logits, axis=2)])
            true_labels.extend(label_ids)

            wandb.log({       
            "Test loss": outputs[0].item(),
            "Validation steps": val_steps
            })
            val_steps += 1

        pred_tags = [tag_values[p_i] for p, l in zip(predictions, true_labels) for p_i, l_i in zip(p, l)]
        valid_tags = [tag_values[l_i] for l in true_labels for l_i in l]
        eval_loss = eval_loss / len(valid_dataloader)
        validation_loss_values.append(eval_loss)
        accuracy = accuracy_score(valid_tags, pred_tags)
        precission = precision_score(valid_tags, pred_tags)
        recall = recall_score(valid_tags, pred_tags)
        f_score = 2 * (precission * recall) / (precission + recall)
        print("Validation Accuracy: {}".format(accuracy))
        print("Validation Precission: {}".format(precission))
        print("Validation Recall: {}".format(recall))
        print("Validation F score: {}".format(f_score))

        wandb.log({       
            "Validation accuracy:": accuracy,
            "Validation precission:": precission,
            "Validation recall": recall,
            "Validation F_score": f_score,
            "Average test loss": eval_loss
            })


    a, p, r ,f = eval_gold(model, dataset, name)
    wandb.log({"Dataset": dataset})
    wandb.log({"Method": name})
    wandb.log({"Acuraccy": a})
    wandb.log({"F_score": f})
    wandb.log({"Precission": p})
    wandb.log({"Recall": r})

    del train_dataloader
    del valid_dataloader
    del tr_inputs
    del tr_labels
    del val_inputs
    del val_labels
    
    gc.collect()
    torch.cuda.empty_cache()

    return model, f_score
    

"""
create a fine-tuned model if model does not exist, otherwise compute metrics on specified dataset
"""
if __name__ == '__main__':

    if len(sys.argv) < 2:
            print("Please specify dataset")
            exit()

    if os.path.isfile('model/pytorch_model.bin'):
        print("Model is created - computing metrics\n")
        model = transformers.BertForTokenClassification.from_pretrained(
        "./model"
        )
        model.to(device)
        model.eval()
        eval_gold(model, sys.argv[1], "fine_tuned_model")

    else: 
        print("Fine tuning model...\n")
        if os.path.isdir(sys.argv[1]+"val/docs") and os.path.isdir(sys.argv[1]+"val/keys"):
            dataset_docs = sys.argv[1] + "train/docs/"
            dataset_keys = sys.argv[1] + "train/keys/"

        elif os.path.isdir(sys.argv[1]+"docs") and os.path.isdir(sys.argv[1]+"keys"):
            print("WARNING: dataset has no evaluation file, metrics will not be computed!")
            dataset_docs = sys.argv[1] + "docs/"
            dataset_keys = sys.argv[1] + "keys/"
        else:
            print("There is no suitable dataset!")
            exit()


        #optimizers = ["adam", "adamw_1", "adamw_2", "sgd"]
        optimizers = ["adamw_2"]
        for i in range(0, len(optimizers)):
            WANDB_API_KEY= "8925984edb6fe740e4626468d3584717a960537f"
            wandb.login(key=WANDB_API_KEY)
            wandb.init(project='keyword extraction', entity='jkrajnak')
            wandb.run.name = "EVAL_model_REG"
            config = wandb.config
            
            docs_names = sorted(os.listdir(dataset_docs), key=len)
            keys_names = sorted(os.listdir(dataset_keys), key=len)

            data = preprocessing(dataset_docs, dataset_keys, docs_names, keys_names)
            input_ids, labels, att_mask = get_tokenized_data(data)
            model, final_fscore = fine_tuning(input_ids, labels, att_mask, optimizers[i], sys.argv[1])

            #directory = "/storage/plzen1/home/jkrajnak/keyword-extraction/model"
            directory = "./model"
            
            if not os.path.exists(directory):
                os.makedirs(directory)

            #save the model weights and its configuration file
            model.save_pretrained(directory)
            print('All files saved')
            

            wandb.log({"Dataset": dataset_docs,
                    "Batch size": BATCH_SIZE,
                    "MAX_LEN": MAX_LEN,
                    "Epoch": (EPOCHS),
                    "Optimizer": optimizers[i]})
            del model
            gc.collect()
            torch.cuda.empty_cache()
            wandb.finish(None)
