from sklearn.feature_extraction.text import TfidfVectorizer
import pandas as pd
import nltk
from nltk.stem import WordNetLemmatizer
nltk.download('wordnet')
wnl = WordNetLemmatizer()

"""
get keywords using TF-IDF

Arguments:
    array_text: list of articles
    lan: datasets language
    n_gram: n gram range
    kw_count: list of keywords count of each article

Returns:
    kw: list of keywords
"""

def get_keywords(text_array, lan, n_gram, kw_count):

    kw = []
    tfidf_vectorizer=TfidfVectorizer(use_idf=True, ngram_range=(1, n_gram)) 
    
    tfidf_vectorizer_vectors=tfidf_vectorizer.fit_transform(text_array)

    for i in range(0, tfidf_vectorizer_vectors.shape[0]):
        vector=tfidf_vectorizer_vectors[i] 

        df = pd.DataFrame(vector.T.todense(), index=tfidf_vectorizer.get_feature_names(), columns=["tfidf"])
        result = df.sort_values(by=["tfidf"],ascending=False)
        keywords = (result.head(len(kw_count[i])).index.tolist())
        lemma = []
        for phrase in keywords:
            if(" " in phrase):
                splitted_phrase = phrase.split(" ")
                lemmatized_output = ' '.join([wnl.lemmatize(w) for w in splitted_phrase])
                lemma.append(lemmatized_output)
            else:
                lemmatized_output = wnl.lemmatize(phrase)
                lemma.append(lemmatized_output)
        kw.append(lemma)
        #print(i)


    return kw

